/*jslint nomen: true, maxlen: 200, indent: 2, maxerr: 100*/
/*global window, document, URL, rJS, RSVP, jIO, tv4 */

(function (window, document, rJS, RSVP, jIO, tv4) {
  "use strict";
  var render_object,
    expandSchema;

  function decodeJsonPointer(_str) {
    // https://tools.ietf.org/html/rfc6901#section-5
    return _str.replace(/~1/g, '/').replace(/~0/g, '~');
  }

  function encodeJsonPointer(_str) {
    // https://tools.ietf.org/html/rfc6901#section-5
    return _str.replace(/~/g, '~0').replace(/\//g, '~1');
  }

  function getBaseUrl(g, path) {
    var base_url,
      key,
      map = g.props.schema_map,
      max_len = 0;
    if (!path) {
      return;
    }
    for (key in map) {
      if (map.hasOwnProperty(key) &&
          path.startsWith(key) &&
          key.length > max_len) {
        base_url = map[key];
        max_len = key.length;
      }
    }
    return base_url;
  }

  function downloadJSON(url) {
    return RSVP.Queue()
      .push(function () {
        return jIO.util.ajax({
          url: url,
          dataType: "json"
        });
      })
      .push(function (evt) {
        return evt.target.response;
      });
  }

  function resolveLocalReference(schema, ref) {
    // 2 here is for #/
    var i, ref_path = ref.substr(2, ref.length),
      parts = ref_path.split("/");
    if (parts.length === 1 && parts[0] === "") {
      // It was uses #/ to reference the entire json so just return it.
      return schema;
    }
    for (i = 0; i < parts.length; i += 1) {
      schema = schema[decodeJsonPointer(parts[i])];
    }
    return schema;
  }

  function schemaPushSchemaPart(schema, schema_path, schema_part) {
    var i,
      k,
      key_list;
    if (schema_path === "/") {
      schema_path = "";
    }
    key_list = schema_path.split("/");
    for (i = 0; i < key_list.length; i += 1) {
      k = decodeJsonPointer(key_list[i]);
      if (i === key_list.length - 1) {
        if (schema_part !== undefined) {
          schema[k] = schema_part;
        } else {
          return schema[k];
        }
      } else {
        if (!schema.hasOwnProperty(k)) {
          schema[k] = {};
        }
        schema = schema[k];
      }
    }
  }

  function getDocumentType(doc) {
    if (doc instanceof Array) {
      return "array";
    }
    return typeof doc;
  }

  function getDocumentSchema(doc) {
    var type = getDocumentType(doc),
      schema = {
        type: type
      };
    if (type === "array") {
      schema.maxItems = 0;
    } else if (type === "object") {
      schema.additionalProperties = false;
    } else {
      schema.readOnly = true;
    }
    return schema;
  }

  function render_selection(schema, json_document) {
    var input = document.createElement("select"),
      option,
      i,
      enum_arr = schema['enum'];
    input.size = 1;
    if (schema.default) {
      if (json_document === undefined) {
        json_document = schema.default;
      }
    } else {
      option = document.createElement("option");
      option.value = "";
      if (json_document === undefined) {
        option.selected = true;
      }
      input.appendChild(option);
    }
    for (i = 0; i < enum_arr.length; i += 1) {
      if (enum_arr.hasOwnProperty(i)) {
        option = document.createElement("option");
        option.value = enum_arr[i];
        option.textContent = enum_arr[i];
        if (enum_arr[i] === json_document) {
          option.selected = true;
        }
        input.appendChild(option);
      }
    }
    return input;
  }

  function render_boolean(schema, json_document) {
    var input,
      schema_for_selection = {
        type: "boolean",
        enum: [true, false]
      };
    // XXX change json_document on open is not correct @bk
    if (json_document === "true") {
      json_document = true;
    }
    if (json_document === "false") {
      json_document = false;
    }
    if (getDocumentType(schema.default) === "boolean") {
      schema_for_selection.default = schema.default;
    }
    input = render_selection(schema_for_selection, json_document);
    return input;
  }

  function render_textarea(json_field, default_value, data_format) {
    var input = document.createElement("textarea");
    if (default_value !== undefined) {
      if (default_value instanceof Array) {
        input.value = default_value.join("\n");
      } else {
        input.value = default_value;
      }
    }
    input["data-format"] = data_format;
    return input;
  }

  function addSubForm(options) {
    var input_element = options.element,
      g = options.gadget,
      property_name,
      parent_path,
      scope;

    scope = "j" + Math.random().toString(36).substr(2, 9);
    if (options.parent_type !== "array") {
      parent_path = options.path;
      property_name = options.property_name;
      if (!property_name) {
        property_name = input_element.value;
      }
      if (!property_name) {
        // XXX notify user
        // you can't create property without property_name
        return RSVP.Queue();
      }
      if (g.props.objects[parent_path].hasOwnProperty(property_name) && g.props.objects[parent_path][property_name] !== "") {
        // XXX notify user
        // you can't create property with existed property_name
        return RSVP.Queue();
      }
      if (input_element) {
        input_element.value = "";
      }
    }

    return g.declareGadget('gadget_json_generated_form.html', {scope: scope})
      .push(function (form_gadget) {
        form_gadget.element.setAttribute("data-gadget-parent-scope",
          g.element.getAttribute("data-gadget-scope"));
        if (options.parent_type !== "array") {
          g.props.objects[parent_path][property_name] = scope;
          form_gadget.element.setAttribute("data-json-parent", parent_path);
          form_gadget.element.setAttribute("data-json-property-name", property_name);
        }
        return form_gadget.renderForm({
          type: options.type,
          schema: options.schema_part,
          schema_path: options.schema_path,
          document: options.default_dict,
          display_label: options.parent_type !== "array",
          scope: scope
        });
      });
  }

  function allOf(g, schema_array, schema_path) {
    return RSVP.Queue()
      .push(function () {
        var i,
          arr = [];
        for (i = 0; i < schema_array.length; i += 1) {
          arr.push(expandSchema(g, schema_array[i], schema_path + '/allOf/' + i.toString()));
        }
        return RSVP.all(arr);
      })
      .push(function (arr) {
        var i,
          x,
          y,
          key,
          next_schema,
          schema,
          schema_item,
          summ_arr;
        for (i = 0; i < arr.length - 1; i += 1) {
          summ_arr = [];
          for (x = 0; x < arr[i].length; x += 1) {
            for (y = 0; y < arr[i + 1].length; y += 1) {
              schema = arr[i][x].schema;
              next_schema = arr[i + 1][y].schema;
              if (schema === true && next_schema === true) {
                schema_item = {
                  schema: true,
                  schema_path: arr[i][x].schema_path
                };
              } else if (schema === false || next_schema === false) {
                schema_item = {
                  schema: false,
                  schema_path: arr[i][x].schema_path
                };
              } else {
                if (schema === true) {
                  schema = {};
                }
                if (next_schema === true) {
                  next_schema = {};
                }
                // copy before change
                schema = JSON.parse(JSON.stringify(schema));
                for (key in next_schema) {
                  if (next_schema.hasOwnProperty(key)) {
                    if (schema.hasOwnProperty(key)) {
                      // XXX need use many many rules for merging
                      schema[key] = next_schema[key];
                    } else {
                      schema[key] = next_schema[key];
                    }
                  }
                }
                schema_item = {
                  schema: schema,
                  schema_path: arr[i][x].schema_path
                };
              }
              summ_arr.push(schema_item);
            }
          }
          arr[i + 1] = summ_arr;
        }
        return arr[arr.length - 1];
      });
  }

  function anyOf(g, schema_array, schema_path) {
    return RSVP.Queue()
      .push(function () {
        var i,
          arr = [];
        for (i = 0; i < schema_array.length; i += 1) {
          arr.push(expandSchema(g, schema_array[i], schema_path + '/anyOf/' + i.toString()));
        }
        return RSVP.all(arr);
      })
      .push(function (arr) {
        var i,
          z,
          schema_arr = [];
        for (i = 0; i < arr.length; i += 1) {
          for (z = 0; z < arr[i].length; z += 1) {
            if (arr[i][z].schema === true) {
              // or(any, restricted, restricted) simplify to any
              return [arr[i][z]];
            }
            schema_arr.push(arr[i][z]);
          }
        }
        return schema_arr;
      });
  }

  expandSchema = function (g, schema, schema_path, ref) {
    // XXX `if then else` construction can be simplify to
    // anyOf(allOf(if_schema, then_schema), else_schema)
    // and realized by existed rails
    if (schema === undefined) {
      schema = true;
    }
    if (schema.anyOf !== undefined) {
      return anyOf(g, schema.anyOf, schema_path);
    }
    if (schema.allOf !== undefined) {
      return allOf(g, schema.allOf, schema_path);
    }
    if (schema.$ref) {
      return g.loadJSONSchema(schema.$ref, schema_path)
        .push(function (schema_part) {
          return expandSchema(g, schema_part, schema_path, schema.$ref);
        });
    }
    return RSVP.Queue()
      .push(function () {
        return [{
          title: ref || schema.title,
          schema: schema,
          schema_path: schema_path
        }];
      });
  };

  function expandProperties(g, properties, schema_path) {
    return RSVP.Queue()
      .push(function () {
        var property_name,
          arr = [];
        function addPropertyName(p_name) {
          return function (schema_array) {
            var i;
            for (i = 0; i < schema_array.length; i += 1) {
              // add propertyName to title
              if (schema_array[i].title && schema_array.length > 1) {
                schema_array[i].title = p_name + ' /' + schema_array[i].title;
              } else {
                schema_array[i].title = p_name;
              }
              // add propertyName to schemaItem
              schema_array[i].property_name = p_name;
            }
            return schema_array;
          };
        }
        for (property_name in properties) {
          if (properties.hasOwnProperty(property_name)) {
            arr.push(
              expandSchema(g, properties[property_name], schema_path +
                  encodeJsonPointer(property_name))
                .push(addPropertyName(property_name))
            );
          }
        }
        return RSVP.all(arr);
      })
      .push(function (arr) {
        var i,
          z,
          schema_arr = [];
        for (i = 0; i < arr.length; i += 1) {
          for (z = 0; z < arr[i].length; z += 1) {
            schema_arr.push(arr[i][z]);
          }
        }
        return schema_arr;
      });
  }

  function checkValidityAndNotifyChange(g) {
    return RSVP.all([
      g.checkValidity(),
      g.notifyChange()
    ]);
  }

  function render_schema_selector(gadget, title, schema_arr, event, rerender) {
    return RSVP.Queue()
      .push(function () {
        var schema_alternatives = [],
          schema_item,
          description,
          i,
          z,
          type;
        function generateItemsForAny(property_name, schema_path) {
          var desc,
            types = [
              "string",
              "number",
              "boolean",
              "array",
              "object",
              "null"
            ],
            ii;
          if (property_name) {
            desc = property_name + " # ";
          } else {
            desc = "";
          }
          for (ii = 0; ii < types.length; ii += 1) {
            schema_alternatives.push({
              title: desc + types[ii],
              value: {
                property_name: property_name,
                schema: { type: types[ii] },
                schema_path: schema_path
              }
            });
          }
        }
        for (i = 0; i < schema_arr.length; i += 1) {
          schema_item = schema_arr[i];
          description = schema_item.title;
          if (schema_item.schema === true) {
            generateItemsForAny(schema_item.property_name, schema_item.schema_path);
          } else if (getDocumentType(schema_item.schema.type) === "array") {
            description = description || schema_item.schema.description;
            for (z = 0; z < schema_item.schema.type.length; z += 1) {
              type = schema_item.schema.type[z];
              schema_alternatives.push({
                title: description + ' # ' + type,
                value: {
                  type: type,
                  property_name: schema_item.property_name,
                  schema_path: schema_item.schema_path,
                  schema: schema_item.schema
                }
              });
            }
          } else {
            description = description ||
              schema_item.schema.type ||
              schema_item.schema.description;
            schema_alternatives.push({
              title: description,
              value: {
                property_name: schema_item.property_name,
                schema_path: schema_item.schema_path,
                schema: schema_item.schema
              }
            });
          }
        }
        return schema_alternatives;
      })
      .push(function (schema_alternatives) {
        var scope = 's' + Math.random().toString(36).substr(2, 9);
        if (schema_alternatives.length > 1) {
          return gadget.declareGadget("gadget_html5_select.html", {scope: scope})
            .push(function (g) {
              return RSVP.Queue()
                .push(function () {
                  var x,
                    item_list = [[title, title]],
                    item;
                  if (rerender) {
                    return rerender(g, schema_alternatives);
                  }
                  for (x = 0; x < schema_alternatives.length; x += 1) {
                    item = schema_alternatives[x];
                    item_list.push([item.title, x]);
                  }
                  return {
                    name: scope,
                    editable: true,
                    hidden: item_list.length === 0,
                    value: item_list[0][1],
                    item_list: item_list
                  };
                })
                .push(function (render_options) {
                  gadget.props.add_custom_data[scope] = {
                    element: g.element,
                    event: function () {
                      return g.getContent()
                        .push(function (value) {
                          return event(schema_alternatives[value[scope]].value);
                        })
                        .push(function () {
                          return checkValidityAndNotifyChange(gadget);
                        })
                        .push(function () {
                          if (rerender) {
                            return rerender(g, schema_alternatives);
                          }
                          return render_options;
                        })
                        .push(function (render_options) {
                          return g.render(render_options);
                        });
                    },
                    rerender: function () {
                      return RSVP.Queue()
                        .push(function () {
                          if (rerender) {
                            return rerender(g, schema_alternatives);
                          }
                          return render_options;
                        })
                        .push(function (render_options) {
                          return g.render(render_options);
                        });
                    }
                  };
                  return g.render(render_options);
                })
                //not need if gadget_html5_select.render return element
                .push(function () {
                  return g.element;
                });
            });
        }
        if (schema_alternatives.length === 1) {
          return RSVP.Queue()
            .push(function () {
              var input = document.createElement("button");
              input.setAttribute("class",
                "ui-shadow-inset ui-btn ui-btn-inline ui-corner-all" +
                " ui-btn-icon-notext ui-icon-btn ui-icon-plus ui-input-btn");
              input.type = "button";
              gadget.props.add_buttons.push({
                element: input,
                event: function () {
                  return event(schema_alternatives[0].value)
                    .push(function () {
                      return checkValidityAndNotifyChange(gadget);
                    });
                }
              });
              return input;
            });
        }
        return RSVP.Queue()
          .push(function () {
            return document.createElement("div");
          });
      });
  }

  function render_array(gadget, json_field, default_array, root, path, schema_path) {
    var queue = RSVP.Queue(),
      div,
      div_input,
      input,
      item_schema,
      i,
      maxItems = json_field.maxItems;
    div = document.createElement("div");
    div.setAttribute("class", "subfield");
    div.title = json_field.description;

    div_input = document.createElement("div");
    div_input.setAttribute("class", "input");

    function element_append(child) {
      if (child) {
        input.parentNode.insertBefore(child, input);
      }
    }

    item_schema = json_field.items;
    queue
      .push(function () {
        return expandSchema(gadget, item_schema, schema_path + '/items');
      })
      .push(function (schema_arr) {
        return render_schema_selector(gadget, "add item to array",
          schema_arr, function (value) {
            return addSubForm({
              gadget: gadget,
              parent_type: 'array',
              type: value.type,
              schema_path: value.schema_path,
              schema_part: value.schema
            })
              .push(element_append);
          });
      })
      .push(function (element) {
        input = element;
        // XXX update on every add/delete item
        input.hidden = maxItems !== undefined && default_array.length >= maxItems;
        div_input.appendChild(input);
        div.appendChild(div_input);
        root.appendChild(div);
      });

    if (default_array) {
      for (i = 0; i < default_array.length; i = i + 1) {
        queue
          .push(
            addSubForm.bind(gadget, {
              gadget: gadget,
              parent_type: 'array',
              schema_path: schema_path + '/items',
              schema_part: item_schema,
              default_dict: default_array[i]
            })
          )
          .push(element_append);
      }
    }
    // XXX add failback rendering if default_array not array
    // input = render_textarea(json_field, default_value, "array");
    return queue;
  }

  function render_field(gadget, key, path, json_field, default_value, root, schema_path, options) {
    var type,
      div,
      label,
      div_input,
      span_info,
      error_message,
      input,
      first_path,
      queue = RSVP.Queue();

    options = options || {};
    type = options.type;

    if (path && key) {
      first_path = path + encodeJsonPointer(key);
    } else {
      first_path = "";
    }

    if (json_field === undefined) {
      json_field = getDocumentSchema(default_value);
    }

    if (getDocumentType(json_field.type) === "string") {
      type = json_field.type;
    } // else json_field.type is array so we use type
    if (type === undefined && default_value !== undefined) {
      type = getDocumentType(default_value);
    }

    // XXX bad peace of code
    // i do not sure that type can be computed so
    // but our schema in slapos bad
    if (!type) {
      if (json_field.properties &&
          json_field.required &&
          json_field.required.length > 0) {
        type = "object";
      }
    }

    div = document.createElement("div");
    div.setAttribute("class", "subfield");
    div.title = json_field.description;
    // if (key && !first_path) {
    if (false) {
      // XXX;
      label = document.createElement("input");
      label.value = key;
      gadget.props.property_name_edit = label;
    } else {
      label = document.createElement("label");
      label.textContent = json_field.title || key;
    }
    div.appendChild(label);
    div_input = document.createElement("div");
    div_input.setAttribute("id", gadget.element.getAttribute("data-gadget-scope") + first_path + '/');
    div_input.setAttribute("class", "input");

    if (json_field.enum !== undefined) {
      input = render_selection(json_field, default_value);
    }

    if (type === "boolean") {
      input = render_boolean(json_field, default_value);
    }

    if (!input && ["string", "integer", "number"].indexOf(type) >= 0) {
      if (json_field.contentMediaType === "text/plain") {
        input = render_textarea(json_field, default_value, "string");
      } else {
        input = document.createElement("input");
        if (default_value !== undefined) {
          input.value = default_value;
        }

        if (type === "integer" || type === "number") {
          input.type = "number";
        } else {
          input.type = "text";
          if (json_field.pattern) {
            input.pattern = json_field.pattern;
          }
        }
      }
    }

    if (type === "array") {
      queue = render_array(
        gadget,
        json_field,
        default_value,
        div_input,
        first_path + '/',
        schema_path
      );
      gadget.props.arrays[first_path +  '/'] = div;
    }

    if (type === "object") {
      queue
        .push(function () {
          return render_object(
            gadget,
            json_field,
            default_value,
            div_input,
            first_path + '/',
            schema_path
          );
        });
    }

    if (input) {
      // object and array excluded from
      // gadget.props.inputs not contain values
      gadget.props.inputs.push(input);
      input.name = first_path;
      input.required = options.required;
      // XXX for gui
      //input.setAttribute("class", "slapos-parameter");
      div_input.appendChild(input);
    } else {
      div.setAttribute("data-json-path", first_path + '/');
      div.setAttribute("data-json-type", type);
    }

    if (json_field.info !== undefined) {
      span_info = document.createElement("span");
      span_info.textContent = json_field.info;
      div_input.appendChild(span_info);
    }
    error_message = document.createElement("span");
    error_message.setAttribute("class", "error");
    error_message.hidden = true;
    div_input.appendChild(error_message);
    div.appendChild(div_input);

    return queue
      .push(function () {
        root.appendChild(div);
        return div;
      });
  }

  render_object = function (g, json_field, default_dict, root, path, schema_path) {
    var additionalProperties,
      key,
      required = json_field.required || [],
      used_properties = {},
      queue = RSVP.Queue(),
      selector = {};

    g.props.objects[path] = used_properties;

    function root_append(child) {
      root.appendChild(child);
    }

    function addAdditional(schema) {
      var div,
        div_input,
        input,
        property_name;

      div = document.createElement("div");
      div.setAttribute("class", "subfield");
      div.title = json_field.description;

      div_input = document.createElement("div");
      div_input.setAttribute("class", "input");

      function element_append(child) {
        if (child) {
          // insert additionalProperty before selector
          selector.element.parentNode.insertBefore(child, selector.element);
        }
      }

      input = document.createElement("input");
      input.type = "text";
      div_input.appendChild(input);

      queue
        .push(function () {
          return expandSchema(g, schema, schema_path + '/additionalProperties');
        })
        .push(function (schema_arr) {
          return render_schema_selector(g, "add custom property", schema_arr, function (value) {
            return addSubForm({
              gadget: g,
              element: input,
              path: path,
              schema_path: value.schema_path,
              schema_part: value.schema
            })
              .push(element_append);
          });
        })
        .push(function (input) {
          // XXX update on every add/delete item
          div_input.appendChild(input);
          div.appendChild(div_input);
        });

      for (property_name in default_dict) {
        if (default_dict.hasOwnProperty(property_name) && !used_properties.hasOwnProperty(property_name)) {
          used_properties[property_name] = "";
          queue
            .push(
              addSubForm.bind(g, {
                gadget: g,
                property_name: property_name,
                path: path,
                schema_path: schema_path + '/additionalProperties',
                schema_part: schema,
                default_dict: default_dict[property_name]
              })
            )
            .push(element_append);
        }
      }
      queue.push(function () {
        root.appendChild(div);
      });
    }

    if (default_dict === undefined) {
      default_dict = {};
    }

    for (key in json_field.properties) {
      if (json_field.properties.hasOwnProperty(key)) {
        if (required.indexOf(key) >= 0) {
          used_properties[key] = false;
          if (json_field.properties[key].default !== undefined) {
            json_field.properties[key].info = '(default = ' + json_field.properties[key].default + ')';
          }
          queue
            .push(render_field.bind(g, g, key, path,
                json_field.properties[key], default_dict[key], root, schema_path, {required: true})
              );
        } else if (default_dict.hasOwnProperty(key)) {
          used_properties[key] = "";
          queue
            .push(
              addSubForm.bind(g, {
                gadget: g,
                property_name: key,
                path: path,
                schema_path: schema_path + '/properties/' + encodeJsonPointer(key),
                schema_part: json_field.properties[key],
                default_dict: default_dict[key]
              })
            )
            .push(root_append);
        }
      }
    }

    queue
      .push(function () {
        return expandProperties(g, json_field.properties, schema_path + '/properties/');
      })
      .push(function (schema_arr) {
        return render_schema_selector(g, "add property", schema_arr, function (value) {
          used_properties[value.property_name] = "";
          return addSubForm({
            gadget: g,
            property_name: value.property_name,
            path: path,
            type: value.type,
            schema_path: value.schema_path,
            schema_part: value.schema
          })
            .push(function (element) {
              var s_e = selector.element;
              if (s_e) {
                s_e.parentNode.insertBefore(element, s_e);
              }
            });
        },
          function (gadget_s, schema_alternatives) {
            var x,
              item_list = [["add property", "add property"]],
              item;
            if (schema_alternatives) {
              for (x = 0; x < schema_alternatives.length; x += 1) {
                item = schema_alternatives[x];
                if (!used_properties.hasOwnProperty(item.value.property_name)) {
                  item_list.push([item.title, x]);
                }
              }
              return {
                name: gadget_s.element.getAttribute('data-gadget-scope'),
                editable: true,
                hidden: item_list.length === 0,
                value: item_list[0][1],
                item_list: item_list
              };
            }
          });
      })
      .push(function (element) {
        selector.element = element;
        return root_append(element);
      });

    if (json_field.patternProperties !== undefined) {
      if (json_field.patternProperties['.*'] !== undefined) {
        addAdditional(json_field.patternProperties['.*']);
      }
    }

    if (json_field.additionalProperties === undefined) {
      additionalProperties = true;
    } else {
      additionalProperties = json_field.additionalProperties;
    }
    if (additionalProperties !== false) {
      addAdditional(additionalProperties);
    }

    for (key in default_dict) {
      if (default_dict.hasOwnProperty(key)) {
        if (!used_properties.hasOwnProperty(key)) {
          queue
            .push(
              addSubForm.bind(g, {
                gadget: g,
                property_name: key,
                path: path,
                schema_path: "",
                schema_part: undefined,
                default_dict: default_dict[key]
              })
            )
            .push(root_append);
        }
      }
    }
    return queue;
  };

  function getFormValuesAsJSONDict(g) {
    var multi_level_dict = {"": {}},
      scope,
      options = g.props,
      array,
      path,
      key,
      i,
      len,
      queue = RSVP.Queue();

    function convertOnMultiLevel(d, key, value) {
      var ii,
        kk,
        key_list = key.split("/");
      for (ii = 0; ii < key_list.length; ii += 1) {
        kk = decodeJsonPointer(key_list[ii]);
        if (ii === key_list.length - 1) {
          if (value !== undefined) {
            d[kk] = value;
          } else {
            return d[kk];
          }
        } else {
          if (!d.hasOwnProperty(kk)) {
            d[kk] = {};
          }
          d = d[kk];
        }
      }
    }

    function recursiveGetContent(scope, path) {
      queue
        .push(function () {
          return g.getDeclaredGadget(scope);
        })
        .push(function (gadget) {
          return gadget.getContent();
        })
        .push(function (jdict) {
          convertOnMultiLevel(multi_level_dict, path, jdict);
        });
    }

    function getContentAndPushArray(scope, parent_path) {
      queue
        .push(function () {
          return g.getDeclaredGadget(scope);
        })
        .push(function (gadget) {
          return gadget.getContent();
        })
        .push(function (jdict) {
          var arr = convertOnMultiLevel(multi_level_dict, parent_path);
          if (!(arr instanceof Array)) {
            arr = [];
            convertOnMultiLevel(multi_level_dict, parent_path, arr);
          }
          arr.push(jdict);
        });
    }

    for (path in options.arrays) {
      if (options.arrays.hasOwnProperty(path)) {
        array = options.arrays[path]
          .querySelectorAll("div[data-gadget-parent-scope='" + g.element.getAttribute("data-gadget-scope") + "']");
        len = array.length;
        for (i = 0; i < len; i = i + 1) {
          getContentAndPushArray(
            array[i].getAttribute('data-gadget-scope'),
            // slice remove concluding '/'
            path.slice(0, -1)
          );
        }
      }
    }

    for (path in options.objects) {
      if (options.objects.hasOwnProperty(path)) {
        for (key in options.objects[path]) {
          if (options.objects[path].hasOwnProperty(key)) {
            scope = options.objects[path][key];
            if (scope) {
              recursiveGetContent(scope, path + encodeJsonPointer(key));
            }
          }
        }
      }
    }

    return queue
      .push(function () {
        var json_dict = {},
          k;
        g.props.inputs.forEach(function (input) {
          if (input.name === "" || input.value !== "") {
            if (input.type === 'number') {
              json_dict[input.name] = parseInt(input.value, 10);
            } else if (input.value === "true") {
              json_dict[input.name] = true;
            } else if (input.value === "false") {
              json_dict[input.name] = false;
            } else if (input.tagName === "TEXTAREA") {
              if (input["data-format"] === "string") {
                json_dict[input.name] = input.value;
              } else {
                json_dict[input.name] = input.value.split('\n');
              }
            } else {
              json_dict[input.name] = input.value;
            }
          }
        });
        for (k in json_dict) {
          if (json_dict.hasOwnProperty(k)) {
            convertOnMultiLevel(multi_level_dict, k, json_dict[k]);
          }
        }
        return multi_level_dict[""];
      });
  }

  function getSubGadgetElement(g, scope) {
    return g.element.querySelector("div[data-gadget-scope='" + scope + "']");
  }

  rJS(window)
    .ready(function () {
      var g = this;
      g.props = {};
      g.options = {};
    })
    .declareAcquiredMethod("notifyChange", "notifyChange")
    .declareAcquiredMethod("renameChildrenParent", "renameChildren")
    .allowPublicAcquisition("renameChildren", function (opt_arr, scope) {
      var property_name,
        objects = this.props.objects,
        new_name = opt_arr[0],
        element = getSubGadgetElement(this, scope),
        parent = element.getAttribute('data-json-parent');
      if (objects.hasOwnProperty(parent)) {
        parent = objects[parent];
        if (parent.hasOwnProperty(new_name)) {
          throw new Error("property already exist");
        }
        // XXX validate property if property pattern
        for (property_name in parent) {
          if (parent.hasOwnProperty(property_name) && parent[property_name] === scope) {
            delete parent[property_name];
            parent[new_name] = scope;
            return new_name;
          }
        }
        throw new Error("gadget not found for renaming");
      }
    })
    .declareMethod("rename", function (new_name, event) {
      var g = this,
        name = g.element.getAttribute('data-json-property-name');
      return this.renameChildrenParent(new_name)
        .push(function () {
          return g.element.setAttribute('data-json-property-name', new_name);
        })
        .push(undefined, function (error) {
          // XXX notify user
          event.srcElement.value = name;
          event.srcElement.focus();
        });
    })
    .declareAcquiredMethod("selfRemove", "deleteChildren")
    .allowPublicAcquisition("deleteChildren", function (arr, scope) {
      var g = this,
        key,
        objects = this.props.objects,
        element = getSubGadgetElement(g, scope),
        parent = element.getAttribute("data-json-parent"),
        tasks = [];
      if (objects.hasOwnProperty(parent)) {
        parent = objects[parent];
        for (key in parent) {
          if (parent.hasOwnProperty(key) && parent[key] === scope) {
            delete parent[key];
          }
        }
      }
      element.parentNode.removeChild(element);
      for (key in g.props.add_custom_data) {
        if (g.props.add_custom_data.hasOwnProperty(key)) {
          tasks.push(g.props.add_custom_data[key].rerender());
        }
      }
      tasks.push(checkValidityAndNotifyChange(g));
      return RSVP.Queue()
        .push(function () {
          return RSVP.all(tasks);
        });
    })

    .declareMethod('getElementByPath', function (data_path) {
      var g = this,
        array,
        path,
        scope,
        key,
        slash_count = 0,
        slash_count_next,
        bingo,
        idx,
        options = g.props;
      if (data_path !== "/") {
        for (path in options.arrays) {
          if (options.arrays.hasOwnProperty(path) && data_path.startsWith(path)) {
            array = options.arrays[path]
              .querySelectorAll("div[data-gadget-parent-scope='" + g.element.getAttribute("data-gadget-scope") + "']");
            data_path = data_path.slice(path.length).split("/");
            idx = data_path[0];
            data_path = "/" + data_path.slice(1).join("/");
            bingo = array[idx].getAttribute('data-gadget-scope');
          }
        }
        if (bingo) {
          return g.getDeclaredGadget(bingo)
            .push(function (gadget) {
              return gadget.getElementByPath(data_path);
            });
        }

        for (path in options.objects) {
          if (options.objects.hasOwnProperty(path) && data_path.startsWith(path)) {
            slash_count_next = path.split("/").length - 1;
            if (slash_count_next > slash_count) {
              bingo = path;
              slash_count = slash_count_next;
            }
          }
        }
        if (bingo) {
          path = options.objects[bingo];
          for (key in path) {
            if (path.hasOwnProperty(key)) {
              if (data_path.startsWith(bingo + encodeJsonPointer(key))) {
                data_path = data_path.slice(bingo.length + encodeJsonPointer(key).length);
                if (!data_path) {
                  data_path = "/";
                }
                scope = path[key];
              }
            }
          }
        }
        if (scope) {
          return g.getDeclaredGadget(scope)
            .push(function (gadget) {
              return gadget.getElementByPath(data_path);
            });
        }
      }
      return RSVP.Queue()
        .push(function () {
          return document.getElementById(
            g.element.getAttribute("data-gadget-scope") + data_path
          );
        });
    })
    .declareAcquiredMethod("notifyValid", "notifyValid")
    .declareAcquiredMethod("notifyInvalid", "notifyInvalid")
    .declareAcquiredMethod("checkValidityParent", "checkValidityTop")
    .allowPublicAcquisition("checkValidityTop", function (arr) {
      return this.checkValidity(arr[0]);
    })
    .declareMethod('checkValidity', function (json_document) {
      // XXX need use local schema and local json document
      // in every subgadget to take into account user anyOf choice
      // and so more precisely point to issue
      var g = this;
      if (!g.props.toplevel) {
        return g.checkValidityParent(json_document);
      }
      return RSVP.Queue()
        .push(function () {
          if (json_document === undefined) {
            return getFormValuesAsJSONDict(g);
          }
          return json_document;
        })
        .push(function (json_d) {
          return tv4.validateMultiple(json_d, g.props.schema[""]);
        })
        .push(function (validation) {
          var index,
            tasks = [];

          g.element.querySelectorAll(".error").forEach(function (error_message) {
            error_message.textContent = "";
            error_message.hidden = true;
          });

          g.element.querySelectorAll("div.error-input").forEach(function (div) {
            div.setAttribute("class", "");
          });
          if (validation.valid) {
            return g.notifyValid()
              .push(function () {
                return false;
              });
          }
          function print_error(message) {
            return function (element) {
              var id = element.id,
                error_message;
              element.setAttribute("class", "error-input");
              error_message = element.querySelector("#" + id.replace("/", "\\/") + " > .error");
              error_message.textContent = message;
              error_message.hidden = false;
            };
          }
          for (index in validation.errors) {
            if (validation.errors.hasOwnProperty(index)) {
              tasks.push(
                g.getElementByPath(validation.errors[index].dataPath)
                  .push(print_error(validation.errors[index].message))
              );
            }
          }

          for (index in validation.missing) {
            if (validation.missing.hasOwnProperty(index)) {
              tasks.push(
                g.getElementByPath(validation.missing[index].dataPath)
                  .push(print_error(validation.missing[index].message))
              );
            }
          }
          return RSVP.Queue()
            .push(function () {
              return RSVP.all(tasks);
            })
            .push(g.notifyInvalid.bind(g))
            .push(function () {
              return false;
            });
        });
    })

    .allowPublicAcquisition("notifyValid", function (arr, sub_scope) {
      return true;
    })
    .allowPublicAcquisition("notifyChange", function (arr, sub_scope) {
      var g = this,
        opt = arr[0],
        event_object;
      event_object = g.props.add_custom_data[sub_scope];
      if (event_object && opt.type === "change") {
        return event_object.event();
      }
      return g.notifyChange();
    })
    .declareMethod('renderForm', function (options) {
      var g = this,
        property_name = g.element.getAttribute('data-json-property-name'),
        schema = options.schema,
        delete_button,
        root;
      g.props.inputs = [];
      g.props.add_buttons = [];
      g.props.add_custom_data = {};
      g.props.arrays = {};
      g.props.objects = {};
      g.props.path = options.path; // self gadget scope
      if (!property_name || !options.display_label) {
        property_name = "";
      }
      root = g.element.querySelector('[data-json-path="/"]');
      if (!root) {
        root = g.element;
      }
      while (root.firstChild) {
        root.removeChild(root.firstChild);
      }
      if (!g.props.toplevel) {
        delete_button = document.createElement("button");
        delete_button.setAttribute("class",
          "ui-shadow-inset ui-btn ui-btn-inline ui-corner-all" +
          " ui-btn-icon-notext ui-icon-btn ui-icon-trash-o ui-input-btn");
        delete_button.type = "button";
        delete_button.name = options.path;
        g.props.delete_button = delete_button;
        root.appendChild(delete_button);
      }
      return render_field(g, property_name, "", schema,
        options.document, root, options.schema_path, {type: options.type})
        .push(function () {
          g.listenEvents();
          return g.element;
        });
    })

    .declareMethod("loadJSONSchema", function (url, path) {
      var g = this,
        protocol,
        download_url,
        base_url,
        hash;
      if (!g.props.toplevel) {
        return g.loadJSONSchemaParent(url, path);
      }
      // XXX need use $id
      base_url = getBaseUrl(g, path);
      if (!base_url) {
        // allow relative link
        base_url = window.location;
      }
      if (!path) {
        path = "/";
      }
      url = new URL(url, base_url);
      protocol = url.protocol;
      if (protocol === "http:" || protocol === "https:") {
        if (window.location.protocol !==  protocol) {
          throw new Error("You cannot mixed http and https calls");
        }
      }
      download_url = url.origin + url.pathname;
      hash = url.hash;
      url = url.href;
      return RSVP.Queue()
        .push(function () {
          if (g.props.schema_cache.hasOwnProperty(download_url)) {
            return g.props.schema_cache[download_url];
          }
          return downloadJSON(download_url)
            .push(function (json) {
              g.props.schema_cache[download_url] = json;
              return json;
            });
        })
        .push(function (json) {
          g.props.schema_map[path] = url;
          return resolveLocalReference(json, hash);
        })
        .push(function (schema_part) {
          // console.log(path);
          schemaPushSchemaPart(g.props.schema, path, schema_part);
          // console.log(g.props.schema[""]);
          return schema_part;
        });
    })
    .declareAcquiredMethod("loadJSONSchemaParent", "loadJSONSchemaTop")
    .allowPublicAcquisition("loadJSONSchemaTop", function (arr) {
      return this.loadJSONSchema(arr[0], arr[1]);
    })
    .declareMethod('render', function (options) {
      var g = this;
      g.props.toplevel = true;
      // contain map of current normalized schema
      // json pointer and corresponding url
      // it's need for schema uri computation
      g.props.schema = {};
      g.props.schema_map = {};
      g.props.schema_cache = {};
      return RSVP.Queue()
        .push(function () {
          if (options.schema) {
            return options.schema;
          }
          var schema_url = options.schema_url ||
            (options.value && options.value.$schema);
          if (schema_url) {
            return g.loadJSONSchema(schema_url);
          }
          return {};
        })
        .push(function (schema) {
          g.options.schema = schema;
          return g.renderForm({
            schema: schema,
            schema_path: "",
            document: options.value
          });
        })
        .push(function () {
          return g.checkValidity();
        })
        .push(function () {
          return g;
        });
    })

    .onEvent('click', function (evt) {
      if (evt.target === this.props.delete_button) {
        return this.selfRemove(evt);
      }

      var button_list = this.props.add_buttons,
        i;
      for (i = 0; i < button_list.length; i = i + 1) {
        if (evt.target === button_list[i].element) {
          return button_list[i].event(evt);
        }
      }
    })

    .onEvent('change', function (evt) {
      if (evt.target === this.props.property_name_edit) {
        return this.rename(this.props.property_name_edit.value, evt);
      }

      var field_list = this.props.inputs,
        i;
      for (i = 0; i < field_list.length; i = i + 1) {
        if (evt.target === field_list[i]) {
          return checkValidityAndNotifyChange(this);
        }
      }
    })
    .declareJob('listenEvents', function () {
      // XXX Disable
      return;
    })

    .declareMethod('getContent', function () {
      var g = this;
      return getFormValuesAsJSONDict(g);
    });

}(window, document, rJS, RSVP, jIO, tv4));